#! /usr/bin/env python3

"""
``senzing_grpc.szdiagnostic.SzDiagnosticGrpc`` is a `gRPC`_ implementation
of the `senzing.szdiagnostic.SzDiagnostic`_ interface.

.. _gRPC: https://grpc.io
.. _senzing.szdiagnostic.SzDiagnostic: https://garage.senzing.com/sz-sdk-python/senzing.html#module-senzing.szdiagnostic
"""

# pylint: disable=E1101

from types import TracebackType
from typing import Any, Dict, Optional, Type, Union

import grpc
from senzing import SzDiagnostic
from senzing_grpc_protobuf import szdiagnostic_pb2, szdiagnostic_pb2_grpc

from .szhelpers import catch_sdk_exceptions, new_exception

# Metadata

__all__ = ["SzDiagnosticGrpc"]
__version__ = "0.0.1"  # See https://www.python.org/dev/peps/pep-0396/
__date__ = "2025-01-10"
__updated__ = "2025-01-16"

SENZING_PRODUCT_ID = (
    "5052"  # See https://github.com/senzing-garage/knowledge-base/blob/main/lists/senzing-component-ids.md
)

# -----------------------------------------------------------------------------
# SzDiagnosticGrpc class
# -----------------------------------------------------------------------------


class SzDiagnosticGrpc(SzDiagnostic):
    """
    SzDiagnostic module access library over gRPC.
    """

    # -------------------------------------------------------------------------
    # Python dunder/magic methods
    # -------------------------------------------------------------------------

    def __init__(
        self,
        grpc_channel: grpc.Channel,
    ) -> None:
        """
        Constructor

        For return value of -> None, see https://peps.python.org/pep-0484/#the-meaning-of-annotations
        """

        # pylint: disable=W0613

        self.channel = grpc_channel
        self.stub = szdiagnostic_pb2_grpc.SzDiagnosticStub(self.channel)

    def __enter__(
        self,
    ) -> Any:  # TODO: Replace "Any" with "Self" once python 3.11 is lowest supported python version.
        """Context Manager method."""
        return self

    def __exit__(
        self,
        exc_type: Union[Type[BaseException], None],
        exc_val: Union[BaseException, None],
        exc_tb: Union[TracebackType, None],
    ) -> None:
        """Context Manager method."""

    # -------------------------------------------------------------------------
    # SzDiagnostic methods
    # -------------------------------------------------------------------------

    @catch_sdk_exceptions
    def check_repository_performance(self, seconds_to_run: int) -> str:
        try:
            request = szdiagnostic_pb2.CheckRepositoryPerformanceRequest(seconds_to_run=seconds_to_run)  # type: ignore[unused-ignore]
            response = self.stub.CheckRepositoryPerformance(request)
            return str(response.result)
        except Exception as err:
            raise new_exception(err) from err

    @catch_sdk_exceptions
    def get_repository_info(self) -> str:
        try:
            request = szdiagnostic_pb2.GetRepositoryInfoRequest()  # type: ignore[unused-ignore]
            response = self.stub.GetRepositoryInfo(request)
            return str(response.result)
        except Exception as err:
            raise new_exception(err) from err

    @catch_sdk_exceptions
    def get_feature(self, feature_id: int) -> str:
        """TODO: Add get_feature()"""
        _ = feature_id
        try:
            request = szdiagnostic_pb2.GetFeatureRequest(feature_id=feature_id)  # type: ignore[unused-ignore]
            response = self.stub.GetFeature(request)
            return str(response.result)
        except Exception as err:
            raise new_exception(err) from err

    def purge_repository(self) -> None:
        """Null function in the sz-sdk-python-grpc implementation."""

    # -------------------------------------------------------------------------
    # Non-public SzDiagnostic methods
    # -------------------------------------------------------------------------

    def _destroy(self) -> None:
        """Null function in the sz-sdk-python-grpc implementation."""

    def initialize(
        self,
        instance_name: str,
        settings: Union[str, Dict[Any, Any]],
        config_id: Optional[int] = None,
        verbose_logging: int = 0,
    ) -> None:
        """Null function in the sz-sdk-python-grpc implementation."""
        _ = instance_name
        _ = settings
        _ = config_id
        _ = verbose_logging

    def reinitialize(self, config_id: int) -> None:
        try:
            request = szdiagnostic_pb2.ReinitializeRequest(config_id=config_id)  # type: ignore[unused-ignore]
            self.stub.Reinitialize(request)
        except Exception as err:
            raise new_exception(err) from err



#! /usr/bin/env python3

"""
``senzing_grpc.szconfigmanager.SzConfigManagerGrpc`` is a `gRPC`_ implementation
of the `senzing.szconfigmanager.SzConfigManager`_ interface.

.. _gRPC: https://grpc.io
.. _senzing.szconfigmanager.SzConfigManager: https://garage.senzing.com/sz-sdk-python/senzing.html#module-senzing.szconfigmanager
"""

# pylint: disable=E1101

from types import TracebackType
from typing import Any, Dict, Type, Union

import grpc
from senzing import SzConfig, SzConfigManager
from senzing_grpc_protobuf import szconfigmanager_pb2, szconfigmanager_pb2_grpc

from .szconfig import SzConfigGrpc
from .szhelpers import as_str, catch_sdk_exceptions, new_exception

# Metadata

__all__ = ["SzConfigManagerGrpc"]
__version__ = "0.0.1"  # See https://www.python.org/dev/peps/pep-0396/
__date__ = "2025-01-10"
__updated__ = "2025-01-16"

SENZING_PRODUCT_ID = (
    "5051"  # See https://github.com/senzing-garage/knowledge-base/blob/main/lists/senzing-component-ids.md
)

# -----------------------------------------------------------------------------
# SzConfigManagerGrpc class
# -----------------------------------------------------------------------------


class SzConfigManagerGrpc(SzConfigManager):
    """
    SzConfigManager module access library over gRPC.
    """

    # -------------------------------------------------------------------------
    # Python dunder/magic methods
    # -------------------------------------------------------------------------

    def __init__(
        self,
        grpc_channel: grpc.Channel,
    ) -> None:
        """
        Constructor

        For return value of -> None, see https://peps.python.org/pep-0484/#the-meaning-of-annotations
        """

        self.channel = grpc_channel
        self.stub = szconfigmanager_pb2_grpc.SzConfigManagerStub(self.channel)

    def __enter__(
        self,
    ) -> Any:  # TODO: Replace "Any" with "Self" once python 3.11 is lowest supported python version.
        """Context Manager method."""
        return self

    def __exit__(
        self,
        exc_type: Union[Type[BaseException], None],
        exc_val: Union[BaseException, None],
        exc_tb: Union[TracebackType, None],
    ) -> None:
        """Context Manager method."""

    # -------------------------------------------------------------------------
    # SzConfigManager methods
    # -------------------------------------------------------------------------

    @catch_sdk_exceptions
    def create_config_from_config_id(self, config_id: int) -> SzConfig:
        try:
            request = szconfigmanager_pb2.GetConfigRequest(config_id=config_id)  # type: ignore[unused-ignore]
            response = self.stub.GetConfig(request)
            config_definition = str(response.result)
            result = SzConfigGrpc(self.channel)
            result.import_config_definition(config_definition)
            return result
        except Exception as err:
            raise new_exception(err) from err

    @catch_sdk_exceptions
    def create_config_from_string(self, config_definition: str) -> SzConfig:
        try:
            result = SzConfigGrpc(self.channel)
            result.import_config_definition(config_definition)
            result.verify_config_definition(config_definition)
            return result
        except Exception as err:
            raise new_exception(err) from err

    @catch_sdk_exceptions
    def create_config_from_template(self) -> SzConfig:
        try:
            request = szconfigmanager_pb2.GetTemplateConfigRequest()  # type: ignore[unused-ignore]
            response = self.stub.GetTemplateConfig(request)
            config_definition = str(response.result)
            result = SzConfigGrpc(self.channel)
            result.import_config_definition(config_definition)
            return result
        except Exception as err:
            raise new_exception(err) from err

    @catch_sdk_exceptions
    def get_config_registry(self) -> str:
        try:
            request = szconfigmanager_pb2.GetConfigRegistryRequest()  # type: ignore[unused-ignore]
            response = self.stub.GetConfigRegistry(request)
            return str(response.result)
        except Exception as err:
            raise new_exception(err) from err

    @catch_sdk_exceptions
    def get_default_config_id(self) -> int:
        try:
            request = szconfigmanager_pb2.GetDefaultConfigIdRequest()  # type: ignore[unused-ignore]
            response = self.stub.GetDefaultConfigId(request)
            return int(response.result)
        except Exception as err:
            raise new_exception(err) from err

    @catch_sdk_exceptions
    def register_config(
        self,
        config_definition: str,
        config_comment: str,
    ) -> int:
        try:
            request = szconfigmanager_pb2.RegisterConfigRequest(  # type: ignore[unused-ignore]
                config_definition=as_str(config_definition),
                config_comment=as_str(config_comment),
            )
            response = self.stub.RegisterConfig(request)
            return int(response.result)
        except Exception as err:
            raise new_exception(err) from err

    @catch_sdk_exceptions
    def replace_default_config_id(self, current_default_config_id: int, new_default_config_id: int) -> None:
        try:
            request = szconfigmanager_pb2.ReplaceDefaultConfigIdRequest(  # type: ignore[unused-ignore]
                current_default_config_id=current_default_config_id,
                new_default_config_id=new_default_config_id,
            )
            self.stub.ReplaceDefaultConfigId(request)
        except Exception as err:
            raise new_exception(err) from err

    @catch_sdk_exceptions
    def set_default_config(self, config_definition: str, config_comment: str) -> int:
        try:
            request = szconfigmanager_pb2.SetDefaultConfigRequest(  # type: ignore[unused-ignore]
                config_definition=as_str(config_definition),
                config_comment=as_str(config_comment),
            )
            response = self.stub.SetDefaultConfig(request)
            return int(response.result)
        except Exception as err:
            raise new_exception(err) from err

    @catch_sdk_exceptions
    def set_default_config_id(self, config_id: int) -> None:
        try:
            request = szconfigmanager_pb2.SetDefaultConfigIdRequest(  # type: ignore[unused-ignore]
                config_id=config_id,
            )
            self.stub.SetDefaultConfigId(request)
        except Exception as err:
            raise new_exception(err) from err

    # -------------------------------------------------------------------------
    # Non-public SzConfigManagerCore methods
    # -------------------------------------------------------------------------

    def _destroy(self) -> None:
        """Null function in the sz-sdk-python-grpc implementation."""

    def initialize(
        self,
        instance_name: str,
        settings: Union[str, Dict[Any, Any]],
        verbose_logging: int = 0,
    ) -> None:
        """Null function in the sz-sdk-python-grpc implementation."""
        _ = instance_name
        _ = settings
        _ = verbose_logging



"""
TODO: szhelpers.py
"""

import json
from collections.abc import Callable
from contextlib import suppress
from functools import wraps
from typing import Any, Dict, TypeVar, Union
from typing import cast as typing_cast

import grpc
from senzing import ENGINE_EXCEPTION_MAP, SzError, SzSdkError

# Metadata

__version__ = "0.0.1"  # See https://www.python.org/dev/peps/pep-0396/
__date__ = "2025-01-10"
__updated__ = "2025-01-16"


_F = TypeVar("_F", bound=Callable[..., Any])

# -----------------------------------------------------------------------------
# Helpers for working with parameters
# -----------------------------------------------------------------------------


def as_str(candidate_value: Union[str, Dict[Any, Any]]) -> str:
    """
    Given a string or dict, return a str.

    Args:
        candidate_value (Union[str, Dict[Any, Any]]): _description_

    Returns:
        str: The string representation of the candidate_value
    """
    if isinstance(candidate_value, dict):
        return json.dumps(candidate_value)
    return candidate_value


# -----------------------------------------------------------------------------
# Helpers for working with errors
# -----------------------------------------------------------------------------


def get_senzing_error_code(error_text: str) -> int:
    """
    Given an exception string, find the exception code.

    :meta private:
    """
    if len(error_text) == 0:
        return 0
    exception_message_splits = error_text.split("|", 1)
    try:
        result = int(exception_message_splits[0].strip().lstrip("SENZ"))
    except ValueError:
        print(f"ERROR: Could not parse error text '{error_text}'")
        result = 9999
    return result


def new_exception(initial_exception: Exception) -> Exception:
    """
    Given an exception, determine which Senzing exception is is.

    Args:
        initial_exception (Exception): An unknown Exception

    Returns:
        Exception: Either a SzError or the original exception.
    """

    result = initial_exception

    if isinstance(initial_exception, grpc.RpcError):

        details = initial_exception.details()  # type: ignore[unused-ignore]

        if details:

            # Find JSON string.

            start_of_json = details.find("{")

            if start_of_json > 0:
                details = details[start_of_json:]

            # Parse JSON.

            details_dict = {}
            try:
                details_dict = json.loads(details)
            except Exception:  # pylint: disable=W0718
                return result

            errors_reason = extract_reason(details_dict)

            # errors_reason = details_dict.get("reason", "")
            senzing_error_code = get_senzing_error_code(errors_reason)
            senzing_error_class = ENGINE_EXCEPTION_MAP.get(senzing_error_code, SzError)
            result = senzing_error_class(details)

    return result


def extract_reason(candidate_dict: Dict[Any, Any]) -> str:
    if "reason" in candidate_dict:
        return str(candidate_dict.get("reason"))
    if "error" in candidate_dict:
        next_dict = candidate_dict.get("error")
        if isinstance(next_dict, dict):
            return extract_reason(next_dict)
    return ""


def catch_sdk_exceptions(func_to_decorate: _F) -> _F:
    """
    The Python SDK methods convert Python types to ctypes and utilize helper functions. If incorrect types/values are
    used standard library exceptions are raised not SzError exceptions as the Senzing library hasn't been called
    yet. Raise the original Python exception type and append information to identify the SDK method called, accepted
    arguments & types and the arguments and types the SDK method received. Also convert ctypes.ArgumentError exceptions
    to TypeError, a user shouldn't need to import ctypes to catch ArgumentError

    :meta private:
    """

    @wraps(func_to_decorate)
    def wrapped_func(*args: Any, **kwargs: Any) -> _F:
        try:
            return typing_cast(_F, func_to_decorate(*args, **kwargs))
        except (TypeError, ValueError) as err:
            # Get wrapped function annotation, remove unwanted keys
            annotations_dict = func_to_decorate.__annotations__
            with suppress(KeyError):
                del annotations_dict["return"]
                del annotations_dict["kwargs"]

            # Get the wrapped function signature names and types and build a string to append to the error message
            func_signature = ", ".join(
                [
                    f"{name}: {type if isinstance(type, str) else type.__name__}"
                    for name, type in annotations_dict.items()
                ]
            )

            method_and_signature = f"{func_to_decorate.__module__}.{func_to_decorate.__name__}({func_signature})"
            append_err_msg = f" - expected: {method_and_signature}"

            arg_0 = err.args[0]
            if " missing " in err.args[0] and " required positional argument" in err.args[0]:
                arg_0 = " ".join(err.args[0].split()[1:])
            new_arg_0 = f"calling {method_and_signature}" if not err.args else f"{arg_0}{append_err_msg}"
            err.args = (new_arg_0,) + err.args[1:]

            raise SzSdkError(err) from err

    return typing_cast(_F, wrapped_func)
